from .rule_template import RuleTemplate


class RuleDDF00083(RuleTemplate):
    """
    DDF00083: Within a study version, all id values must be unique.

    Applies to: All
    Attributes: id
    """

    def __init__(self):
        super().__init__(
            "DDF00083",
            RuleTemplate.ERROR,
            "Within a study version, all id values must be unique.",
        )

    def validate(self, config: dict) -> bool:
        # See rule DDF00082 for schema checks
        return True



from typing import Union, Literal
from .api_base_model import ApiBaseModelNameLabelDesc


class SyntaxTemplate(ApiBaseModelNameLabelDesc):
    text: str
    dictionaryId: Union[str, None] = None
    instanceType: Literal["SyntaxTemplate"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .quantity import Quantity
from .agent_administration import AgentAdministration
from .code import Code


class StudyIntervention(ApiBaseModelNameLabelDesc):
    role: Code
    type: Code
    minimumResponseDuration: Union[Quantity, None] = None
    codes: List[Code] = []
    administrations: List[AgentAdministration] = []
    productDesignation: Code
    pharmacologicClass: Union[Code, None] = None
    instanceType: Literal["StudyIntervention"]



from .rule_template import RuleTemplate


class RuleDDF00144(RuleTemplate):
    """
    DDF00144: A study geographic scope type must be specified using the geographic scope type (C207412) DDF codelist.

    Applies to: GeographicScope
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00144",
            RuleTemplate.ERROR,
            "A study geographic scope type must be specified using the geographic scope type (C207412) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "GeographicScope", "type")



from typing import Literal, Union
from .syntax_template import SyntaxTemplate
from .code import Code


class EligibilityCriterion(SyntaxTemplate):
    category: Code
    identifier: str
    nextId: Union[str, None] = None
    previousId: Union[str, None] = None
    contextId: Union[str, None] = None
    instanceType: Literal["EligibilityCriterion"]



from .rule_template import RuleTemplate


class RuleDDF00105(RuleTemplate):
    """
    DDF00105: A scheduled activity/decision instance must only reference an epoch that is defined within the same study design as the scheduled activity/decision instance.

    Applies to: ScheduledActivityInstance, ScheduledDecisionInstance
    Attributes: epoch
    """

    def __init__(self):
        super().__init__(
            "DDF00105",
            RuleTemplate.ERROR,
            "A scheduled activity/decision instance must only reference an epoch that is defined within the same study design as the scheduled activity/decision instance.",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("ScheduledActivityInstance")
        items += data.instances_by_klass("ScheduledDecisionInstance")
        for item in items:
            if "epochId" in item:
                epoch = data.instance_by_id(item["epochId"])
                if epoch:
                    item_parent = data.parent_by_klass(item["id"], "StudyDesign")
                    epoch_parent = data.parent_by_klass(epoch["id"], "StudyDesign")
                    if item_parent["id"] != epoch_parent["id"]:
                        self._add_failure(
                            "Epoch defined in a different study design",
                            item["instanceType"],
                            "epochId",
                            data.path_by_id(item["id"]),
                        )
        return self._result()



from .rule_template import RuleTemplate


class RuleDDF00104(RuleTemplate):
    """
    DDF00104: A timing's relative to/from property must be specified using the Timing Relative To From Value Set Terminology (C201265) SDTM codelist.

    Applies to: Timing
    Attributes: relativeToFrom
    """

    def __init__(self):
        super().__init__(
            "DDF00104",
            RuleTemplate.ERROR,
            "A timing's relative to/from property must be specified using the Timing Relative To From Value Set Terminology (C201265) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Timing", "relativeToFrom")



from d4k_sel.errors import Errors
from usdm3.rules.library.rule_template import ValidationLocation


class RulesValidationResults:
    def __init__(self):
        self._items = {}

    def add_success(self, rule: str):
        self._items[rule] = {
            "status": "Success",
            "errors": None,
            "exception": None,
        }

    def add_failure(self, rule: str, errors: Errors):
        self._items[rule] = {
            "status": "Failure",
            "errors": [],
            "exception": None,
        }
        for error in errors._items:
            item = error.to_dict()
            self._items[rule]["errors"].append(item)

    def add_exception(self, rule: str, exception: Exception):
        self._items[rule] = {
            "status": "Exception",
            "errors": None,
            "exception": str(exception),
        }

    def add_not_implemented(self, rule: str):
        self._items[rule] = {
            "status": "Not Implemented",
            "errors": None,
            "exception": None,
        }

    def count(self):
        return len(self._items)

    def passed(self):
        return all(item["status"] == "Success" for item in self._items.values())

    def passed_or_not_implemented(self):
        return all(
            item["status"] == "Success" or item["status"] == "Not Implemented"
            for item in self._items.values()
        )

    def to_dict(self) -> list[dict]:
        if len(self._items) == 0:
            return []
        else:
            rows = []
            location = ValidationLocation(
                rule="", rule_text="", klass="", attribute="", path=""
            )
            for rule, item in self._items.items():
                if item["errors"]:
                    for error in item["errors"]:
                        row = {
                            "rule_id": rule,
                            "status": item["status"],
                            "exception": item["exception"],
                        }
                        row.update(self._flatten_error(error))
                        rows.append(row)
                else:
                    row = {
                        "rule_id": rule,
                        "status": item["status"],
                        "exception": item["exception"],
                    }
                    row.update(
                        self._flatten_error(
                            {"message": "", "level": "", "location": location.to_dict()}
                        )
                    )
                    rows.append(row)
            rows.sort(key=lambda x: x["rule_id"])
            return rows

    def _flatten_error(self, error: dict) -> dict:
        for key, value in error["location"].items():
            error[key] = value
        error.pop("location")
        return error



from typing import Literal, Union
from .api_base_model import ApiBaseModel
from .quantity import Quantity


class AdministrationDuration(ApiBaseModel):
    quantity: Union[Quantity, None] = None
    description: str
    durationWillVary: bool
    reasonDurationWillVary: str
    instanceType: Literal["AdministrationDuration"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc


class Condition(ApiBaseModelNameLabelDesc):
    text: str
    dictionaryId: Union[str, None] = None
    contextIds: List[str] = []
    appliesToIds: List[str] = []
    instanceType: Literal["Condition"]



from typing import Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc


class BiomedicalConceptSurrogate(ApiBaseModelNameLabelDesc):
    reference: Union[str, None] = None
    instanceType: Literal["BiomedicalConceptSurrogate"]



from .rule_template import RuleTemplate


class RuleDDF00155(RuleTemplate):
    """
    DDF00155: For CDISC codelist references (where the code system is 'http://www.cdisc.org'), the code system version must be a valid CDISC terminology release date in ISO 8601 date format.

    Applies to: Code
    Attributes: codeSystemVersion
    """

    VERSION_LIST = [
        "2020-03-27",
        "2020-05-08",
        "2020-06-26",
        "2020-09-25",
        "2020-11-06",
        "2020-12-18",
        "2021-03-26",
        "2021-06-25",
        "2021-09-24",
        "2021-12-17",
        "2022-03-25",
        "2022-06-24",
        "2022-09-30",
        "2022-12-16",
        "2023-03-31",
        "2023-06-30",
        "2023-09-29",
        "2023-12-15",
        "2024-03-29",
        "2024-09-27",
    ]

    def __init__(self):
        super().__init__(
            "DDF00155",
            RuleTemplate.ERROR,
            "For CDISC codelist references (where the code system is 'http://www.cdisc.org'), the code system version must be a valid CDISC terminology release date in ISO 8601 date format.",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("Code")
        for item in items:
            if "codeSystem" in item and item["codeSystem"] == "http://www.cdisc.org":
                if "codeSystemVersion" not in item:
                    self._add_failure(
                        "Missing codeSystemVersion",
                        "Code",
                        "codeSystemVersion",
                        data.path_by_id(item["id"]),
                    )
                else:
                    if item["codeSystemVersion"] not in self.VERSION_LIST:
                        self._add_failure(
                            "Invalid codeSystemVersion",
                            "Code",
                            "codeSystemVersion",
                            data.path_by_id(item["id"]),
                        )
        return self._result()



from .rule_template import RuleTemplate


class RuleDDF00146(RuleTemplate):
    """
    DDF00146: A study title type must be specified using the study title type (C207419) DDF codelist.

    Applies to: StudyTitle
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00146",
            RuleTemplate.ERROR,
            "A study title type must be specified using the study title type (C207419) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "StudyTitle", "type")



from .rule_template import RuleTemplate


class RuleDDF00025(RuleTemplate):
    """
    DDF00025: A window must not be defined for an anchor timing (i.e., type is \"Fixed Reference\").

    Applies to: Timing
    Attributes: windowLabel, windowLower, windowUpper
    """

    def __init__(self):
        super().__init__(
            "DDF00025",
            RuleTemplate.ERROR,
            'A window must not be defined for an anchor timing (i.e., type is "Fixed Reference").',
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("Timing")
        for item in items:
            if item["type"]["decode"] == "Fixed Reference":
                if "windowLower" in item:
                    if item["windowLower"] is not None:
                        self._add_failure(
                            "Window lower defined for anchor timing",
                            "Timing",
                            "windowLower",
                            data.path_by_id(item["id"]),
                        )
                if "windowUpper" in item:
                    if item["windowUpper"] is not None:
                        self._add_failure(
                            "Window upper defined for anchor timing",
                            "Timing",
                            "windowUpper",
                            data.path_by_id(item["id"]),
                        )
        return self._result()



from typing import List, Literal
from .api_base_model import ApiBaseModel


class StudyCell(ApiBaseModel):
    armId: str
    epochId: str
    elementIds: List[str] = []
    instanceType: Literal["StudyCell"]



class IdManager:
    def __init__(self, classes: list[str]):
        self._classes = classes
        self._id_index = {}
        self.clear()

    def clear(self):
        for klass in self._classes:
            name = klass if isinstance(klass, str) else klass.__name__
            self._id_index[name] = 0

    def build_id(self, klass):
        klass_name = klass if isinstance(klass, str) else str(klass.__name__)
        self._id_index[klass_name] += 1
        return f"{klass_name}_{self._id_index[klass_name]}"



from .rule_template import RuleTemplate


class RuleDDF00067(RuleTemplate):
    """
    DDF00067: A study cell must refer to at least one element.

    Applies to: StudyCell
    Attributes: elements
    """

    def __init__(self):
        super().__init__(
            "DDF00067",
            RuleTemplate.ERROR,
            "A study cell must refer to at least one element.",
        )



import os
import yaml


class Missing:
    def __init__(self):
        f = open(os.path.join(os.path.dirname(__file__), "missing_ct.yaml"))
        self._missing_ct = yaml.load(f, Loader=yaml.FullLoader)

    def code_lists(self):
        for response in self._missing_ct:
            yield response



from typing import List, Literal
from .api_base_model import ApiBaseModel
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent


class StudyProtocolDocumentVersion(ApiBaseModel):
    protocolVersion: str
    protocolStatus: Code
    dateValues: List[GovernanceDate] = []
    contents: List[NarrativeContent] = []
    childIds: List[str] = []
    instanceType: Literal["StudyProtocolDocumentVersion"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModel
from .study_identifier import StudyIdentifier
from .alias_code import AliasCode
from .code import Code
from .study_design import StudyDesign
from .governance_date import GovernanceDate
from .study_amendment import StudyAmendment
from .study_title import StudyTitle


class StudyVersion(ApiBaseModel):
    versionIdentifier: str
    rationale: str
    studyType: Union[Code, None] = None
    studyPhase: Union[AliasCode, None] = None
    documentVersionId: Union[str, None] = None
    dateValues: List[GovernanceDate] = []
    amendments: List[StudyAmendment] = []
    businessTherapeuticAreas: List[Code] = []
    studyIdentifiers: List[StudyIdentifier] = []
    studyDesigns: List[StudyDesign] = []
    titles: List[StudyTitle]
    instanceType: Literal["StudyVersion"]



from typing import List, Literal, Union
from pydantic import Field
from .api_base_model import ApiBaseModel
from .study_protocol_document import StudyProtocolDocument
from .study_version import StudyVersion
from uuid import UUID


class Study(ApiBaseModel):
    id: Union[UUID, None] = None
    name: str = Field(min_length=1)
    description: Union[str, None] = None
    label: Union[str, None] = None
    versions: List[StudyVersion] = []
    documentedBy: Union[StudyProtocolDocument, None] = None
    instanceType: Literal["Study"]



from d4k_sel import ErrorLocation


class SchemaErrorLocation(ErrorLocation):
    def __init__(self, path: str, instance):
        self.path = path
        self.instance = instance

    def to_dict(self):
        return {"path": self.path, "instance": self.instance}

    def __str__(self):
        return f"[{self.path}, {self.instance}]"



import pathlib
from usdm3.rules.rules_validation import RulesValidation3
from usdm3.rules.rules_validation_results import RulesValidationResults
from usdm3.minimum.minimum import Minimum
from usdm3.api.wrapper import Wrapper


class USDM3:
    def __init__(self):
        self.root = self._root_path()
        self.validator = RulesValidation3(self.root, "usdm3.rules.library")

    def validate(self, file_path: str) -> RulesValidationResults:
        return self.validator.validate(file_path)

    def minimum(self, study_name: str, sponsor_id: str, version: str) -> Wrapper:
        return Minimum.minimum(study_name, sponsor_id, version)

    def _root_path(self) -> str:
        return pathlib.Path(__file__).parent.resolve()



from typing import List, Literal, Union
from .activity import Activity
from .api_base_model import ApiBaseModelNameLabelDesc
from .alias_code import AliasCode
from .biomedical_concept import BiomedicalConcept
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .code import Code
from .encounter import Encounter
from .study_cell import StudyCell
from .indication import Indication
from .study_intervention import StudyIntervention
from .study_arm import StudyArm
from .study_epoch import StudyEpoch
from .study_element import StudyElement
from .population_definition import StudyDesignPopulation
from .objective import Objective
from .schedule_timeline import ScheduleTimeline
from .estimand import Estimand
from .syntax_template_dictionary import SyntaxTemplateDictionary
from .masking import Masking
from .condition import Condition
from .organization import ResearchOrganization


class StudyDesign(ApiBaseModelNameLabelDesc):
    trialIntentTypes: List[Code] = []
    trialTypes: List[Code] = []
    therapeuticAreas: List[Code] = []
    characteristics: List[Code] = []
    interventionModel: Code
    encounters: List[Encounter] = []
    activities: List[Activity] = []
    biomedicalConcepts: List[BiomedicalConcept] = []
    bcCategories: List[BiomedicalConceptCategory] = []
    bcSurrogates: List[BiomedicalConceptSurrogate] = []
    arms: List[StudyArm]
    studyCells: List[StudyCell]
    blindingSchema: Union[AliasCode, None] = None
    rationale: str
    epochs: List[StudyEpoch]
    elements: List[StudyElement] = []
    estimands: List[Estimand] = []
    indications: List[Indication] = []
    maskingRoles: List[Masking] = []
    studyInterventions: List[StudyIntervention] = []
    objectives: List[Objective] = []
    population: Union[StudyDesignPopulation, None] = None
    scheduleTimelines: List[ScheduleTimeline] = []
    documentVersionId: Union[str, None] = None
    dictionaries: List[SyntaxTemplateDictionary] = []
    conditions: List[Condition] = []
    organizations: List[ResearchOrganization] = []
    instanceType: Literal["StudyDesign"]



from .rule_template import RuleTemplate


class RuleDDF00149(RuleTemplate):
    """
    DDF00149: A study arm data origin type must be specified according to the extensible data origin type (C188727) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: StudyArm
    Attributes: dataOriginType
    """

    def __init__(self):
        super().__init__(
            "DDF00149",
            RuleTemplate.ERROR,
            "A study arm data origin type must be specified according to the extensible data origin type (C188727) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "StudyArm", "dataOriginType")



from typing import Literal
from .api_base_model import ApiBaseModel
from .organization import Organization


class StudyIdentifier(ApiBaseModel):
    studyIdentifier: str
    studyIdentifierScope: Organization
    instanceType: Literal["StudyIdentifier"]



from .rule_template import RuleTemplate


class RuleDDF00008(RuleTemplate):
    """
    DDF00008: A scheduled activity instance must refer to either a default condition or a timeline exit, but not both.

    Applies to: ScheduledActivityInstance
    Attributes: timelineExit, defaultCondition
    """

    def __init__(self):
        super().__init__(
            "DDF00008",
            RuleTemplate.ERROR,
            "A scheduled activity instance must refer to either a default condition or a timeline exit, but not both.",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("ScheduledActivityInstance")
        for item in items:
            if "timelineExitId" in item and "defaultConditionId" in item:
                timeline_exit = data.instance_by_id(item["timelineExitId"])
                default_condition = data.instance_by_id(item["defaultConditionId"])
                if timeline_exit and default_condition:
                    self._add_failure(
                        "Timeline exit and default condition both exist",
                        item["instanceType"],
                        "timelineExitId and defaultConditionId",
                        data.path_by_id(item["id"]),
                    )
        return self._result()



from .rule_template import RuleTemplate


class RuleDDF00031(RuleTemplate):
    """
    DDF00031: If timing type is not \"Fixed Reference\" then it must point to two scheduled instances (e.g. the relativeFromScheduledInstance and relativeToScheduledInstance attributes must not be missing and must not be equal to each other).

    Applies to: Timing
    Attributes: relativeToScheduledInstance
    """

    def __init__(self):
        super().__init__(
            "DDF00031",
            RuleTemplate.ERROR,
            'If timing type is not "Fixed Reference" then it must point to two scheduled instances (e.g. the relativeFromScheduledInstance and relativeToScheduledInstance attributes must not be missing and must not be equal to each other).',
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("Timing")
        for item in items:
            check = True
            if item["type"]["decode"] != "Fixed Reference":
                if "relativeToScheduledInstanceId" not in item:
                    self._add_failure(
                        "Missing relativeToScheduledInstanceId",
                        "Timing",
                        "relativeToScheduledInstanceId",
                        data.path_by_id(item["id"]),
                    )
                    check = False
                if "relativeFromScheduledInstanceId" not in item:
                    self._add_failure(
                        "Missing relativeFromScheduledInstanceId",
                        "Timing",
                        "relativeFromScheduledInstanceId",
                        data.path_by_id(item["id"]),
                    )
                    check = False
                if (
                    check
                    and item["relativeToScheduledInstanceId"]
                    == item["relativeFromScheduledInstanceId"]
                ):
                    self._add_failure(
                        "relativeToScheduledInstanceId and relativeFromScheduledInstanceId are equal",
                        "Timing",
                        "relativeToScheduledInstanceId and relativeFromScheduledInstanceId",
                        data.path_by_id(item["id"]),
                    )
        return self._result()



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelNameLabel
from .code import Code
from .address import Address
from .study_site import StudySite


class Organization(ApiBaseModelNameLabel):
    organizationType: Code
    identifierScheme: str
    identifier: str
    legalAddress: Union[Address, None] = None
    instanceType: Literal["Organization"]


class ResearchOrganization(Organization):
    manages: List[StudySite]
    instanceType: Literal["ResearchOrganization"]



from .rule_template import RuleTemplate


class RuleDDF00106(RuleTemplate):
    """
    DDF00106: A scheduled activity instance must only reference an encounter that is defined within the same study design as the scheduled activity instance.

    Applies to: ScheduledActivityInstance
    Attributes: encounter
    """

    def __init__(self):
        super().__init__(
            "DDF00106",
            RuleTemplate.ERROR,
            "A scheduled activity instance must only reference an encounter that is defined within the same study design as the scheduled activity instance.",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("ScheduledActivityInstance")
        for item in items:
            if "encounterId" in item:
                encounter = data.instance_by_id(item["encounterId"])
                if encounter:
                    item_parent = data.parent_by_klass(item["id"], "StudyDesign")
                    encounter_parent = data.parent_by_klass(
                        encounter["id"], "StudyDesign"
                    )
                    if item_parent["id"] != encounter_parent["id"]:
                        self._add_failure(
                            "Encounter defined in a different study design",
                            item["instanceType"],
                            "encounterId",
                            data.path_by_id(item["id"]),
                        )
        return self._result()



from .rule_template import RuleTemplate


class RuleDDF00118(RuleTemplate):
    """
    DDF00118: A study design's trial intent types must be specified according to the extensible Trial Intent Type Response (C66736) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: StudyDesign
    Attributes: trialIntentTypes
    """

    def __init__(self):
        super().__init__(
            "DDF00118",
            RuleTemplate.ERROR,
            "A study design's trial intent types must be specified according to the extensible Trial Intent Type Response (C66736) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )



from .rule_template import RuleTemplate


class RuleDDF00045(RuleTemplate):
    def __init__(self):
        super().__init__(
            "DDF00045",
            RuleTemplate.WARNING,
            "At least one attribute must be specified for an address.",
        )

    def validate(self, config: dict):
        data = config["data"]
        addresses = data.instances_by_klass("Address")
        for address in addresses:
            address_valid = False
            for attribute in [
                "text",
                "line",
                "city",
                "district",
                "state",
                "postalCode",
                "country",
            ]:
                if attribute in address and address[attribute]:
                    address_valid = True
            if not address_valid:
                self._add_failure(
                    "No attributes specified for address",
                    "Address",
                    "",
                    data.path_by_id(address["id"]),
                )
        return self._result()



from .rule_template import RuleTemplate


class RuleDDF00009(RuleTemplate):
    """
    DDF00009: Each schedule timeline must contain at least one anchor (fixed time) - i.e., at least one scheduled activity instance that is referenced by a Fixed Reference timing.

    Applies to: Timing
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00009",
            RuleTemplate.ERROR,
            "Each schedule timeline must contain at least one anchor (fixed time) - i.e., at least one scheduled activity instance that is referenced by a Fixed Reference timing.",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("ScheduledTimeline")
        for item in items:
            valid = False
            for timing in item["timings"]:
                if timing["type"]["decode"] == "Fixed Reference":
                    valid = True
            if not valid:
                self._add_failure(
                    "No fixed reference timing",
                    item["instanceType"],
                    "timings",
                    data.path_by_id(item["id"]),
                )
        return self._result()



from .activity import Activity
from .address import Address
from .administration_duration import AdministrationDuration
from .agent_administration import AgentAdministration
from .alias_code import AliasCode
from .analysis_population import AnalysisPopulation
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_property import BiomedicalConceptProperty
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .biomedical_concept import BiomedicalConcept
from .code import Code
from .condition import Condition
from .eligibility_criterion import EligibilityCriterion
from .encounter import Encounter
from .endpoint import Endpoint
from .estimand import Estimand
from .geographic_scope import GeographicScope
from .governance_date import GovernanceDate
from .indication import Indication
from .intercurrent_event import IntercurrentEvent
from .masking import Masking
from .narrative_content import NarrativeContent
from .objective import Objective
from .organization import Organization
from .population_definition import PopulationDefinition
from .procedure import Procedure
from .quantity import Quantity
from .range import Range
from .response_code import ResponseCode
from .schedule_timeline_exit import ScheduleTimelineExit
from .schedule_timeline import ScheduleTimeline
from .scheduled_instance import ScheduledInstance
from .study_amendment import StudyAmendment
from .study_amendment_reason import StudyAmendmentReason
from .study_arm import StudyArm
from .study_cell import StudyCell
from .study_design import StudyDesign
from .study_element import StudyElement
from .study_epoch import StudyEpoch
from .study_identifier import StudyIdentifier
from .study_intervention import StudyIntervention
from .study_protocol_document_version import StudyProtocolDocumentVersion
from .study_protocol_document import StudyProtocolDocument
from .study_site import StudySite
from .study_title import StudyTitle
from .study_version import StudyVersion
from .study import Study
from .syntax_template import SyntaxTemplate
from .syntax_template_dictionary import SyntaxTemplateDictionary
from .timing import Timing
from .transition_rule import TransitionRule
from .wrapper import Wrapper

__all__ = [
    "Activity",
    "Address",
    "AdministrationDuration",
    "AgentAdministration",
    "AliasCode",
    "AnalysisPopulation",
    "BiomedicalConceptCategory",
    "BiomedicalConceptProperty",
    "BiomedicalConceptSurrogate",
    "BiomedicalConcept",
    "Code",
    "Condition",
    "EligibilityCriterion",
    "Encounter",
    "Endpoint",
    "Enrollment",
    "Estimand",
    "GeographicScope",
    "GovernanceDate",
    "Indication",
    "IntercurrentEvent",
    "InvestigationalIntervention",
    "Masking",
    "NarrativeContent",
    "Objective",
    "Organization",
    "PopulationDefinition",
    "Procedure",
    "Quality",
    "Range",
    "ResponseCode",
    "ResearchOrganization",
    "ScheduleTimelineExit",
    "ScheduleTimeline",
    "ScheduledInstance",
    "ScheduledActivityInstance",
    "ScheduledDecisionInstance",
    "StudyAmendment",
    "StudyAmendmentReason",
    "StudyArm",
    "StudyCell",
    "StudyCohort",
    "StudyDesignPopulation",
    "StudyDesign",
    "StudyElement",
    "StudyEpoch",
    "StudyIdentifier",
    "StudyProtocolDocumentVersion",
    "StudyProtocolDocument",
    "StudySite",
    "StudyTitle",
    "StudyVersion",
    "Study",
    "SubjectEnrollment",
    "SyntaxTemplate",
    "SyntaxTemplateDictionary",
    "Timing",
    "TransitionRule",
    "StudyIntervention",
    "Quantity",
    "Wrapper",
]



from typing import List, Literal
from .api_base_model import ApiBaseModelNameLabelDesc
from .study_protocol_document_version import StudyProtocolDocumentVersion


class StudyProtocolDocument(ApiBaseModelNameLabelDesc):
    versions: List[StudyProtocolDocumentVersion] = []
    instanceType: Literal["StudyProtocolDocument"]



class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        name = cls.__name__
        if name not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[name] = instance
        return cls._instances[name]

    # Method to assist with testing.
    @classmethod
    def _clear(cls, the_cls):
        name = the_cls.__name__
        if name in cls._instances:
            del cls._instances[name]



from .rule_template import RuleTemplate


class RuleDDF00029(RuleTemplate):
    """
    DDF00029: An encounter must only reference encounters that are specified within the same study design.

    Applies to: Encounter
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00029",
            RuleTemplate.ERROR,
            "An encounter must only reference encounters that are specified within the same study design.",
        )



from .rule_template import RuleTemplate


class RuleDDF00143(RuleTemplate):
    """
    DDF00143: A study amendment reason must be coded using the study amendment reason (C207415) DDF codelist.

    Applies to: StudyAmendmentReason
    Attributes: code
    """

    def __init__(self):
        super().__init__(
            "DDF00143",
            RuleTemplate.ERROR,
            "A study amendment reason must be coded using the study amendment reason (C207415) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "StudyAmendmentReason", "code")



import json
import enum
import datetime
from typing import Union
from pydantic import BaseModel, Field
from uuid import UUID


def _serialize_as_json(obj):
    if isinstance(obj, enum.Enum):
        return obj.value
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    else:
        return obj.__dict__


class ApiBaseModel(BaseModel):
    id: str = Field(min_length=1)

    def to_json(self):
        return json.dumps(self, default=_serialize_as_json)


class ApiBaseModelAndDesc(ApiBaseModel):
    description: Union[str, None] = None


class ApiBaseModelName(ApiBaseModel):
    name: str = Field(min_length=1)


class ApiBaseModelNameLabelDesc(ApiBaseModelName):
    label: Union[str, None] = None
    description: Union[str, None] = None


class ApiBaseModelNameLabel(ApiBaseModelName):
    label: Union[str, None] = None


class ApiBaseModelNameDesc(ApiBaseModelName):
    description: Union[str, None] = None



import os
import json
from jsonschema import validate, RefResolver, ValidationError


class SchemaValidation:
    class Dummy(ValidationError):
        """Include so ruff doesnot complain about the ValidationError class not being used. Want it to be propagated."""

        pass

    def __init__(self, schema_file_path: str):
        self.schema_file_path = schema_file_path
        with open(schema_file_path, "r") as f:
            self.schema = json.load(f)

        # Create a resolver for handling $ref references in the schema
        schema_path = os.path.abspath(schema_file_path)
        schema_uri = f"file://{schema_path}"
        self.resolver = RefResolver(schema_uri, self.schema)

    def get_component_schema(self, component_name):
        if (
            "components" not in self.schema
            or "schemas" not in self.schema["components"]
        ):
            raise ValueError("Schema does not contain components/schemas section")

        if component_name not in self.schema["components"]["schemas"]:
            raise ValueError(f"Schema component '{component_name}' not found")

        return self.schema["components"]["schemas"][component_name]

    def validate_against_component(self, data, component_name):
        component_schema = self.get_component_schema(component_name)
        validate(instance=data, schema=component_schema, resolver=self.resolver)
        return True

    def validate_file(self, json_file_path, component_name):
        with open(json_file_path, "r") as f:
            data = json.load(f)
        return self.validate_against_component(data, component_name)



import os
from usdm3.ct.cdisc.library_api import LibraryAPI
from usdm3.ct.cdisc.config.config import Config
from usdm3.ct.cdisc.missing.missing import Missing
from usdm3.ct.cdisc.library_cache.library_cache import LibraryCache


class Library:
    """
    A class to manage CDISC controlled terminology (CT) data.

    This class handles loading, caching, and accessing CDISC controlled terminology,
    including code lists and their associated terms. It can load data from a local
    cache file or fetch it from the CDISC API when needed.
    """

    def __init__(self, filepath: str):
        self.filepath = filepath
        self._config = Config(
            os.path.join(self.filepath, "config")
        )  # Configuration for required code lists and mappings
        self._missing = Missing()  # Handler for missing/additional code lists
        self._api = LibraryAPI()  # Interface to CDISC Library API
        self._cache = LibraryCache(
            os.path.join(self.filepath, "library_cache")
        )  # Cache file handler

        # Data structures to store and index controlled terminology
        self._by_code_list = {}  # Maps concept IDs to complete code list data
        self._by_term = {}  # Maps term concept IDs to parent code list IDs
        self._by_submission = {}  # Maps submission values to parent code list IDs
        self._by_pt = {}  # Maps preferred terms to parent code list IDs

    def load(self) -> None:
        if self._cache.exists():
            self._load_ct()  # Load from cache file
        else:
            self._api.refresh()  # Ensure API connection is fresh
            self._get_ct()  # Fetch from API
            self._cache.save(self._by_code_list)  # Cache the results
        self._add_missing_ct()  # Add any additional required terminology

    def klass_and_attribute(self, klass, attribute) -> dict:
        try:
            concept_id = self._config.klass_and_attribute(klass, attribute)
            return self._by_code_list[concept_id]
        except Exception:
            return None

    def _get_ct(self) -> None:
        for item in self._config.required_code_lists():
            response = self._api.code_list(item)
            self._by_code_list[response["conceptId"]] = response
            for item in response["terms"]:
                # Index each term by its various identifiers
                self._check_in_and_add(
                    self._by_term, item["conceptId"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_pt, item["preferredTerm"], response["conceptId"]
                )

    def _load_ct(self) -> None:
        self._by_code_list = self._cache.read()
        for c_code, entry in self._by_code_list.items():
            for item in entry["terms"]:
                # Rebuild indexes from cached data
                self._check_in_and_add(self._by_term, item["conceptId"], c_code)
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], c_code
                )
                self._check_in_and_add(self._by_pt, item["preferredTerm"], c_code)

    def _add_missing_ct(self) -> None:
        for response in self._missing.code_lists():
            self._by_code_list[response["conceptId"]] = response
            for item in response["terms"]:
                # Index the additional terms
                self._check_in_and_add(
                    self._by_term, item["conceptId"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_pt, item["preferredTerm"], response["conceptId"]
                )

    def _check_in_and_add(self, collection: dict, id: str, item: str) -> None:
        if id not in collection:
            collection[id] = []
        collection[id].append(item)



from typing import Literal, Union
from .api_base_model import ApiBaseModel
from .code import Code


class Range(ApiBaseModel):
    minValue: float
    maxValue: float
    unit: Union[Code, None] = None
    isApproximate: bool
    instanceType: Literal["Range"]



from typing import Literal, Union
from .api_base_model import ApiBaseModel
from .code import Code


class Address(ApiBaseModel):
    text: Union[str, None] = None
    line: Union[str, None] = None
    city: Union[str, None] = None
    district: Union[str, None] = None
    state: Union[str, None] = None
    postalCode: Union[str, None] = None
    country: Union[Code, None] = None
    instanceType: Literal["Address"]



from typing import List, Literal
from .api_base_model import ApiBaseModel
from .analysis_population import AnalysisPopulation
from .intercurrent_event import IntercurrentEvent


class Estimand(ApiBaseModel):
    summaryMeasure: str
    analysisPopulation: AnalysisPopulation
    interventionId: str
    variableOfInterestId: str
    intercurrentEvents: List[IntercurrentEvent]
    instanceType: Literal["Estimand"]



from .rule_template import RuleTemplate


class RuleDDF00131(RuleTemplate):
    """
    DDF00131: Referenced items in the narrative content must be available elsewhere in the data model.

    Applies to: NarrativeContent
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "DDF00131",
            RuleTemplate.ERROR,
            "Referenced items in the narrative content must be available elsewhere in the data model.",
        )



from typing import Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .code import Code


class StudyEpoch(ApiBaseModelNameLabelDesc):
    type: Code
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    instanceType: Literal["StudyEpoch"]



import os
import re
import requests


class LibraryAPI:
    API_ROOT = "https://api.library.cdisc.org/api"

    class APIError(Exception):
        """Custom exception for API-related errors"""

        pass

    def __init__(self) -> None:
        self._packages = None
        self._headers = {
            "Content-Type": "application/json",
            "api-key": os.environ.get("CDISC_API_KEY"),
        }

    def refresh(self) -> None:
        self._packages = self._get_packages()

    def code_list(self, c_code: str) -> dict:
        use_list = ["ddfct", "sdtmct", "protocolct"]
        for package in use_list:
            try:
                version = self._packages[package][-1]["effective"]
            except Exception:
                version = None
            if version:
                package_full_name = f"{package}-{version}"
                api_url = self._url(
                    f"/mdr/ct/packages/{package_full_name}/codelists/{c_code}"
                )
                raw = requests.get(api_url, headers=self._headers)
                if raw.status_code == 200:
                    response = raw.json()
                    response.pop("_links", None)
                    return response
        raise self.APIError(
            f"failed to obtain code list from library for {c_code}, "
            f"response: {raw.status_code} {raw.text}"
        )

    def _get_packages(self) -> dict:
        packages = {}
        api_url = self._url("/mdr/ct/packages")
        raw = requests.get(api_url, headers=self._headers)
        if raw.status_code == 200:
            response = raw.json()
            for item in response["_links"]["packages"]:
                name = self._extract_ct_name(item["href"])
                effective_date = self._extract_effective_date(item["title"])
                if name and effective_date:
                    if name not in packages:
                        packages[name] = []
                    packages[name].append(
                        {"effective": effective_date, "url": item["href"]}
                    )
            return packages
        else:
            raise self.APIError(
                f"failed to get packages, response: {raw.status_code} {raw.text}"
            )

    def _extract_ct_name(self, url: str) -> str:
        match = re.search(r"([a-zA-Z]+)-\d{4}-\d{2}-\d{2}$", url)
        if match:
            return match.group(1)
        return None

    def _extract_effective_date(self, title: str) -> str:
        match = re.search(r"Effective (\d{4}-\d{2}-\d{2})$", title)
        if match:
            return match.group(1)
        return None

    def _url(self, relative_url: str) -> str:
        # print(f"{self.API_ROOT} + {relative_url}")
        return f"{self.API_ROOT}{relative_url}"



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelName


class NarrativeContent(ApiBaseModelName):
    sectionNumber: str
    sectionTitle: str
    text: Union[str, None] = None
    childIds: List[str] = []
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    instanceType: Literal["NarrativeContent"]



from typing import Literal, Union
from .api_base_model import ApiBaseModel
from .alias_code import AliasCode


class Quantity(ApiBaseModel):
    value: float
    unit: Union[AliasCode, None] = None
    instanceType: Literal["Quantity"]



import json
from d4k_sel.errors import Errors
from d4k_sel import ErrorLocation


class DataStoreErrorLocation(ErrorLocation):
    def __init__(self, path: str, missing: str, extra: str):
        self.path = path
        self.missing = missing
        self.extra = extra

    def to_dict(self):
        return {"path": self.path, "missing": self.missing}

    def __str__(self):
        extra = f", {self.extra}" if self.extra else ""
        return f"[{self.path}, missing '{self.missing}' attribute{extra}]"


class DecompositionError(Exception):
    def __init__(self, error: DataStoreErrorLocation):
        self.error = error

    def __str__(self):
        return f"error decomposing the '.json' file, missing {self.error.missing} at {self.error.path}"


class DataStore:
    def __init__(self, filename: str):
        self._klasses = {}
        self._ids = {}
        self._parent = {}
        self._path = {}
        self.filename = filename
        self.data = None
        self.errors = Errors

    def decompose(self):
        self.data = self._load_data()
        self._decompose(self.data, None, "")

    def instance_by_id(self, id: str) -> dict:
        if id not in self._ids:
            return None
        return self._ids[id]

    def path_by_id(self, id: str) -> str:
        if id not in self._path:
            return None
        return self._path[id]

    def instances_by_klass(self, klass: str) -> list:
        if klass not in self._klasses:
            return []
        return list(self._klasses[klass].values())

    def parent_by_klass(self, id: str, klass: str) -> dict:
        found = False
        if id not in self._ids:
            return None
        instance = self._ids[id]
        while not found:
            if instance["instanceType"] == klass:
                found = True
            else:
                instance = self._parent[instance["id"]]
        return instance

    def _decompose(self, data, parent, path, instance_index=None) -> None:
        if isinstance(data, dict):
            path = self._add_klass_instance(data, parent, path, instance_index)
            for key, value in data.items():
                if isinstance(value, dict):
                    self._decompose(value, data, path)
                elif isinstance(value, list):
                    for index, item in enumerate(value):
                        self._decompose(item, data, path, index)

    def _add_klass_instance(self, data, parent, path, instance_index) -> None:
        id, klass = self._check_id_klass(parent, data, path)
        path = self._update_path(path, data, instance_index)
        if klass not in self._klasses:
            self._klasses[klass] = {}
        self._klasses[klass][id] = data
        self._ids[id] = data
        self._path[id] = path
        self._parent[id] = parent
        return path

    def _load_data(self) -> dict:
        with open(self.filename, "r") as file:
            return json.load(file)

    def _update_path(self, path: str, data: dict, instance_index: int) -> str:
        path = path + "." + data["instanceType"] if "instanceType" in data else "$"
        path = path + f"[{instance_index}]" if instance_index is not None else path
        return path

    def _check_id_klass(self, parent: dict, data: dict, path: str) -> None:
        id, error = self._check_id(parent, data, path)
        if error:
            raise DecompositionError(error)
        klass, error = self._check_instance_type(parent, data, path)
        if error:
            raise DecompositionError(error)
        return id, klass

    def _check_id(self, parent: dict, data: dict, path: str) -> None:
        if parent:
            if "id" in data:
                return data["id"], None
            else:
                klass = data["instanceType"] if "instanceType" in data else ""
                return None, DataStoreErrorLocation(
                    path, "id", f"with instanceType '{klass}'"
                )
        else:
            return "$root", None

    def _check_instance_type(self, parent: dict, data: dict, path: str) -> None:
        if parent:
            if "instanceType" in data:
                return data["instanceType"], None
            else:
                id = data["id"] if "id" in data else ""
                return None, DataStoreErrorLocation(
                    path, "instanceType", f"with id '{id}'"
                )
        else:
            return "Wrapper", None



from .rule_template import RuleTemplate


class RuleDDF00140(RuleTemplate):
    """
    DDF00140: An organization type must be specified according to the extensible organization type (C188724) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: Organization, ResearchOrganization
    Attributes: organizationType
    """

    def __init__(self):
        super().__init__(
            "DDF00140",
            RuleTemplate.ERROR,
            "An organization type must be specified according to the extensible organization type (C188724) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Organization", "organizationType")



import os
import sys
import inspect
import importlib
from pathlib import Path
from typing import List, Type
from usdm3.rules.library.rule_template import RuleTemplate
from usdm3.data_store.data_store import DataStore, DecompositionError
from usdm3.ct.cdisc.library import Library as CTLibrary
from usdm3.rules.rules_validation_results import RulesValidationResults
from usdm3.base.singleton import Singleton


class RulesValidation3(metaclass=Singleton):
    def __init__(self, root_path: str, package_name: str):
        self.rules_validation = RulesValidationEngine(root_path, package_name)

    def validate(self, filename: str):
        return self.rules_validation.validate_rules(filename)


class RulesValidationEngine:
    def __init__(self, root_path: str, package_name: str):
        self.root_path = root_path
        print(f"LIBRARY: {root_path}, {package_name}")
        self.library_path = os.path.join(self.root_path, "rules/library")
        self.ct_path = os.path.join(self.root_path, "ct/cdisc")
        # print(f"PATHS: {self.root_path}, {self.library_path}, {self.ct_path}")
        self.package_name = package_name
        self.rules: List[Type[RuleTemplate]] = []
        self._load_rules()

    def validate_rules(self, filename: str) -> RulesValidationResults:
        data_store, e = self._data_store(filename)
        if data_store:
            ct = CTLibrary(self.ct_path)
            ct.load()
            config = {"data": data_store, "ct": ct}
            results = self._execute_rules(config)
        else:
            results = RulesValidationResults()
            results.add_exception("Decomposition", e)
        return results

    def _data_store(self, filename: str) -> DataStore:
        try:
            data_store = DataStore(filename)
            data_store.decompose()
            return data_store, None
        except DecompositionError as e:
            return None, e

    def _load_rules(self) -> None:
        # Iterate through all .py files in the library directory
        for file in Path(self.library_path).glob("rule_*.py"):
            if file.name.startswith("rule_ddf") and file.name.endswith(".py"):
                try:
                    # Create module name from file name
                    module_name = f"{self.package_name}.{file.stem}"

                    # Load module using absolute path
                    spec = importlib.util.spec_from_file_location(
                        module_name, str(file)
                    )
                    if spec is None or spec.loader is None:
                        continue

                    module = importlib.util.module_from_spec(spec)
                    sys.modules[spec.name] = module
                    spec.loader.exec_module(module)

                    for name, obj in inspect.getmembers(module):
                        if (
                            inspect.isclass(obj)
                            and issubclass(obj, RuleTemplate)
                            and obj != RuleTemplate
                            and obj.__module__.startswith(self.package_name)
                        ):
                            try:
                                self.rules.append(obj)
                            except Exception:
                                continue

                except Exception:
                    continue

    def _execute_rules(self, config: dict) -> RulesValidationResults:
        results = RulesValidationResults()
        for rule_class in self.rules:
            try:
                # Execute the rule
                rule: RuleTemplate = rule_class()
                passed = rule.validate(config)
                if passed:
                    results.add_success(rule._rule)
                else:
                    results.add_failure(rule._rule, rule.errors())
            except NotImplementedError:
                # Rule not implemented yet
                results.add_not_implemented(rule._rule)
            except Exception as e:
                # print(f"RULE: {rule._rule} exception: {e}")
                # print(traceback.format_exc())
                results.add_exception(rule._rule, e)
        return results



from typing import List, Literal
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelNameLabel
from .biomedical_concept_property import BiomedicalConceptProperty


class BiomedicalConcept(ApiBaseModelNameLabel):
    synonyms: List[str] = []
    reference: str
    properties: List[BiomedicalConceptProperty] = []
    code: AliasCode
    instanceType: Literal["BiomedicalConcept"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .alias_code import AliasCode


class BiomedicalConceptCategory(ApiBaseModelNameLabelDesc):
    childIds: List[str] = []
    memberIds: List[str] = []
    code: Union[AliasCode, None] = None
    instanceType: Literal["BiomedicalConceptCategory"]
